/************************************************************************
 * inner_control.c                                                      *
 * Mijan Huq                                                            *
 *                                                                      *
 ************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "path.h"    /* should be included in mesh.h */
#include "option.h"  /* should be included in mesh.h */
#include "mesh.h"    /* mesh data structures */
#include "util.h" 
#include "blackScholes.h"
#include "inner.h"

/* Routine that computes inner controls nu defined in paper page 18.  
   Sets up nu(t,X...,X...)  Is this the correct way? 
   This is needed for computing alpha and beta */
void ComputeInnerControl(int whichInnerControl, meshData *theMesh, 
                         double *control, int *istar, int t)
{
  int  path, underlier, whichunderlier;
  double max, deltat,pmax, pmin, min;

  /* compute delta t */
  deltat = 1.0; 

  switch ( whichInnerControl){

  case 1:   /* Control 1 in paper */

      /* Loop over all paths */
      for(path = 0 ; path < theMesh->option.pData.paths; path++){

        /* Find the maximum price at current time */
	    max = 0.0;
        for(underlier=0; underlier < theMesh->option.pData.underliers; 
            underlier++){
          pmax = max;
          max=dMax(max,theMesh->option.pData.price[t][path][underlier]);
          if (max != pmax) {
            istar[path] = underlier;
          }
        } /* End loop over underliers */
        /* Now the control is max(argmax(S^i_t+1 - K)) */
        control[path] = exp(-deltat*(theMesh->option.riskFreeRate )) * \
          dMax(max - theMesh->option.strikePrice, 0.0);

      } /* End loop over paths */
    break;
  case 2:   /* Control 2 in paper */
      /* Loop over all paths */
      for(path = 0 ; path < theMesh->option.pData.paths; path++){

        /* Find the maximum price at time+1 */
           
	    max = 0.0;
        for(underlier=0; underlier < theMesh->option.pData.underliers; 
            underlier++){
          pmax = max;
          max=dMax(max,theMesh->option.pData.price[t][path][underlier]);
          if (max != pmax) {
            istar[path] = underlier;
          }
        } /* End loop over underliers */
         
        /* Now the control is (argmax(S^i_t+1)) */
        control[path] = max;

      } /* End loop over paths */
    break;
    
  case 3:   /* Control 3 in paper */
    perror("ComputeInnerControl: Control %d not yet implmented");
    exit (1);
  } /* end switch over inner control */
   
} /* end ComputeInnerControl */


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* Routine that computes a Cumulative normal dist based on Hull 243 */
/* Six decimal accurate */
double CumNormDist(double x){
  double a1, a2, a3, a4, a5, gam, k, value,dNdx;
  a1 = 0.319381530;
  a2 = -0.356563782;
  a3 = 1.781477937;
  a4 = -1.821255978;
  a5 = 1.330274429;
  gam = 0.2316419;
  k = 1.0 / (1.0 + gam * x);
  dNdx = 1.0/sqrt(2.0 * M_PI)*exp(-x*x*0.5);

  if(x >= 0.0){
    value = 1.0 - dNdx * \
      (a1*k + a2*pow(k,2) + a3*pow(k,3) + a4*pow(k,4)  + a5*pow(k,5));
  }else{
    value = 1.0 - CumNormDist(-x); 
  }
  /*
    printf("%lf %lf %lf\n",x,value,  dNdx);
  */
  return (value);
}


/*
   Routine that computes independent estimate of nu 
*/
double nuIndEst(int whichInnerControl, double S, double r, double dividend,\
                double sigma, double T, double t, double strikePrice){

  double nuest, d1, d2;

  switch ( whichInnerControl){

  case 1:   /* Control 1 in paper */
    /* Evaluate Black-Scholes formula with dividends (Merton's model) */
    d1 = (log(S/strikePrice) + ((r-dividend) + sigma * sigma *0.5)*(T-t)) \
      /sigma/sqrt(T-t);
          
    d2 = (log(S/strikePrice) + ((r-dividend) - sigma * sigma *0.5)*(T-t)) \
      /sigma/sqrt(T-t);

    nuest = exp(-dividend*(T-t)) * S * CumNormDist(d1) - \
      strikePrice*exp(-r *(T-t))*CumNormDist(d2);
    break;
  case 2:   /* Control 2 in paper */
    nuest = S * exp((r-dividend)*(T-t));
    break;
  case 3:  /* Control 3 in paper */
    /* The following is obtained from Stulz 1982 page 163, eqn 11 */
    break;      
  } /* end switch */
  return (nuest);
}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/* Routine that uses control variates to modify given values of \hat{Q} */
void innerControlVariate(int whichInnerControl, 
                         meshData *mesh_in, meshData *mesh_out)
{
  int Npaths, Ntimes;
  double *Icontrol, **value, **valueIfHeld; 
  int *imax;

  int i,j,t, b;

  double alpha, beta, denominator, A,B,C,D,E,F, sigma, nuest;


  /* Set up a few local parameters for convenience */
  Npaths = mesh_in->option.pData.paths;
  Ntimes = mesh_in->option.pData.times;
  b = mesh_in->meshPaths; 

  /* Allocate memory */
  Icontrol = (double *)malloc(Npaths * sizeof(double ));
  value = (double **)malloc(Ntimes * sizeof(double *));
  valueIfHeld = (double **)malloc(Ntimes * sizeof(double *));
  Icontrol = (double *)malloc(Npaths * sizeof(double ));
  imax = (int *)malloc(Npaths * sizeof(int ));
  if(Icontrol == NULL || imax == NULL){
    perror("innerControlVariate: Could not allocate memory");
    exit(1);
  }
  for(i=0;i<Ntimes; i++){
      value[i] = (double *)malloc(Npaths * sizeof(double));
      valueIfHeld[i] = (double *)malloc(Npaths * sizeof(double));
  }


  /* loop over times */
  printf("weight at 0,0,0 %lf\n",mesh_in->weights[0][0][0]);
  for(t=mesh_in->option.pData.times-2; t>=0; t--){
    printf("time = %d\n",t);
    for(i=0; i < mesh_in->option.pData.paths; i++){

	  /* Compute inner controls at each mesh point for time t*/
	  ComputeInnerControl(whichInnerControl, mesh_in, Icontrol, imax,t);

      /* Compute beta from weighted least squares fits */
      beta = 0.0;

      /* compute <Q> where <Q> := E[Q(t+1,S_{t+1} | S_t = X_t(i)] */
      A = 0.0;   
      for(j=0;j<b; j++){
        A +=  mesh_in->weights[t][i][j] * mesh_in->value[t+1][j];
      }
      
      /* Compute Sum_j w(t,i,j) */ 
      B = 0.0;
      for(j=0;j<b; j++){
        B +=  mesh_in->weights[t][i][j];
      }
     
      /* Compute <nu> */
      C = 0.0;
      for(j=0;j<b; j++){
        C +=  mesh_in->weights[t][i][j]  * Icontrol[j]; 
      }

      /* Compute <Q*nu> */
      D = 0.0;
      for(j=0;j<b; j++){
	    D +=  mesh_in->weights[t][i][j]  * Icontrol[j] *
	  mesh_in->value[t+1][j]; 
      }

      E = C;

      /* Compute <nu*nu> */
      F = 0.0;
      for(j=0;j<b; j++){ /* term F in handwritten notes */
	        F += mesh_in->weights[t][i][j] * Icontrol[j]*
	  Icontrol[j]; 
      }
      
      beta =    (D * B - E*A)/(F*B - E*C);
                        
      /* Compute alpha */
      alpha = (A - beta * C) / B;

      /* Compute independent source for nu */
      
      sigma = sqrt(mesh_in->option.pData.covariance[0][0]); 

      nuest =  nuIndEst(whichInnerControl, 
			mesh_in->option.pData.price[t][i][imax[i]],
                        mesh_in->option.riskFreeRate, 
			mesh_in->option.pData.dividendRate[imax[i]], 
			sigma, 
			/*
			  (double)mesh_in->option.pData.times, 
			  (double)t,   
			*/
			1.0, 0.0,
                        mesh_in->option.strikePrice);

      /* Then we update Q(t,i)... */
      valueIfHeld[t][i] = alpha + beta * nuest; 


      /* Now recheck which is larger, valueIfHeld or payoff */
      if(mesh_out->option.payoff[t][i] >  mesh_out->valueIfHeld[t][i]){
        value[t][i] = mesh_out->option.payoff[t][i];
      }else{
        value[t][i] = mesh_out->valueIfHeld[t][i];
      }


    } /* loop over i */
  } /* End loop over time here */
 printf("Mesh estimate after inner control= %lf\n",valueIfHeld[0][0]);
  free(Icontrol);
  

} /* end innerControlVariate */
